/*****************************************************************
 * Copyright (c) 2017-2019 Texas Instruments and others
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Paul Gingrich - Initial API and implementation
 *****************************************************************/
var gc = gc || {};
gc.databind = gc.databind || {};
gc.databind.internal = gc.databind.internal || {};
gc.databind.internal.reg = gc.databind.internal.reg || {};

(function()
{
    /**
     * parseCoreKey returns a numeric value for the core number specified by coreKey string.
     * if coreKey === 'all': returns -1
     * if coreKey === 'active':
     * Must be called with this = the register model to be used.
     */
    var parseCoreKey = function(coreKey) {
        var result = 0;
        if (!isNaN(coreKey)) {
            result = +coreKey;
        } else {
            switch ("" + coreKey) {
                case "active":
                    // TODO: not sure how to assign the selected-core attribute of ti-model-reg to $selectedCore
                    break;
                case "all":
                    result = -1;
                    break;
            }
        }
        return result;
    };

    var typeParser = /\s*(unsigned\s|signed\s)?\s*(int|q(\d+))\s*/i;

    var onRegParentChangedListener = function()
    {
        var regAllBind = this.registerAllBind;
        if (!regAllBind) {
            regAllBind = this.parentBind;
        }
        if (!regAllBind) return;
        var value = regAllBind.getValueFor(this.coreIndex);
        if ((this._maxValue !== undefined) && (value >= this._maxValue))
        {
            // convert to signed
            value = value - (this._maxValue + 1);
        }
        if ((this._q !== undefined) && !isNaN(value))
        {
            value = value / (Math.pow(2, this._q));
        }

        this.updateValue(value, undefined, true);
    };

    var getBitfield = function(value) {
        value = gc.utils.getBitfieldValue(value, this._startBit, this._stopBit);

        if ((this._maxValue !== undefined) && (value >= this._maxValue)) {
            // convert to signed
            value = value - (this._maxValue + 1);
        }
        if ((this._q !== undefined) && !isNaN(value)) {
            value = value / (Math.pow(2, this._q));
        }
        return value;
    };

    var onFieldParentChangedListener = function()
    {
        var value = this.parentBind.getValue();
        if (!Array.isArray(value)) {
            value = getBitfield.call(this,value);
        } else {
            value = value.slice();
            for (var i=0; i < value.length; i++) {
                value[i] = getBitfield.call(this,value[i]);
            }
        }
        this.updateValue(value, undefined, true);
    };

    var onStatusListener = function(status)
    {
        if (!this.fCachedStatus)
        {
            // relay parent binding status (null or error), unless we are reporting our own error message.
            this.setStatus(status);
        }
    };

    gc.databind.internal.reg.RegisterBind = function(uri, model, refreshIntervalProvider, symbolData, registerAllBind)
    {
        gc.databind.AbstractAsyncBindValue.call(this, 'number');
        this.uri = uri;
        this.parentModel = model;
        if (model.isDeviceArray) {
            // TODO: is the following line correct??
            try {
                gc.databind.internal.reg.RegisterBind.prototype.excludeFromStorageProviderData = true;
            } catch(ex1){}
            this.parentBind = registerAllBind;
            this.coreIndex = 0;
            var parentChangedListener = new gc.databind.IChangedListener();
            parentChangedListener.onValueChanged = onRegParentChangedListener.bind(this);
            parentChangedListener.onStatusChanged = onStatusListener.bind(this);
            if (registerAllBind) {
                registerAllBind.addChangedListener(parentChangedListener);
                registerAllBind.addStatusListener(parentChangedListener);

                this.dispose = function()
                {
                    registerAllBind.removeChangedListener(parentChangedListener);
                    registerAllBind.removeStatusListener(parentChangedListener);
                };
            }
        }

        var that = this;
        that._onFailure = function(errMsg)
        {
            if (that.parentModel.isConnected())
            {
                that.reportErrorStatus(errMsg);
            }
            var callback = that._callback;
            that._callback = undefined;
            if (callback) {
                callback(that.fCachedValue);  // don't record a new value, keep the same value as before.
            }
        };

        that._onSuccess = function(result)
        {
            // clear errors on succesfull read
            that.reportCriticalError(null);

            return result;
        };

        this.setRefreshIntervalProvider(refreshIntervalProvider);


        this.updateRegisterInfo(symbolData);
    };

    gc.databind.internal.reg.RegisterBind.prototype = new gc.databind.AbstractAsyncBindValue('number');

    // TODO: excludeFromStorageProviderData should be set to true if isDeviceArray = true.
    // gc.databind.internal.reg.RegisterBind.prototype.excludeFromStorageProviderData = true;

    //[... Methods that are used only if model.isDeviceArray is true
    gc.databind.internal.reg.RegisterBind.prototype.setIndex = function(index) {
        if (this.parentModel && this.parentModel.isDeviceArray && !isNaN(index)) {
            var oldValue = this.parentBind.getValueFor(this.coreIndex);
            this.coreIndex = index;
            var newValue = this.parentBind.getValueFor(index);
            // call updateValue so we don't commit the new value to the target
            this.updateValue(newValue);
            // notify listeners
            if (newValue !== undefined) {
                this.notifyChangedListeners(oldValue, newValue, null);
            }
        }
    };

    gc.databind.internal.reg.RegisterBind.prototype.getIndex = function(){
        return this.coreIndex;
    };

    gc.databind.internal.reg.RegisterBind.prototype.onFirstDataReceivedListenerAdded = function()
    {
        if (this.parentBind && this.parentModel && this.parentModel.isDeviceArray) {
            this.parentBind.addStreamingListener(this);
        } else {
            gc.databind.AbstractAsyncBindValue.prototype.onFirstDataReceivedListenerAdded.call(this);
        }
    };


    /**
     * Method called when the last streaming listener is removed from the list.
     * Derived classes can override this method to be notified for this event.
     */
    gc.databind.internal.reg.RegisterBind.prototype.onLastDataReceivedListenerRemoved = function()
    {
        if (this.parentBind && this.parentModel && this.parentModel.isDeviceArray) {
            this.parentBind.removeStreamingListener(this);
        } else {
            gc.databind.AbstractAsyncBindValue.prototype.onLastDataReceivedListenerRemoved.call(this);
        }
    };

    gc.databind.internal.reg.RegisterBind.prototype.onDataReceived = function(newValue)
    {
        if (this.parentBind && this.parentModel && this.parentModel.isDeviceArray) {
            this.notifyDataReceivedListeners(this.fCachedValue, this.coreIndex);
        } else {
            try {
                gc.databind.StreamingDataBind.prototype.onDataReceived.call(this, newValue);
            } catch(ex) {
                console.error("RegisterBind.onDataReceived: could not forward to StreamingDataBind. ex="+ex);
            }
        }
    };

    gc.databind.internal.reg.RegisterBind.prototype.isStale = function()
    {
        if (this.parentBind && this.parentModel && this.parentModel.isDeviceArray) {
            return this.parentBind.isStale();
        } else {
            return gc.databind.AbstractBindValue.prototype.isStale.call(this);
        }
    };

    gc.databind.internal.reg.RegisterBind.prototype.addStaleListener = function(listener)
    {
        if (this.parentBind && this.parentModel && this.parentModel.isDeviceArray) {
            this.parentBind.addStaleListener(listener);
        } else {
            gc.databind.AbstractBindValue.prototype.addStaleListener.call(this, listener);
        }
    };

    gc.databind.internal.reg.RegisterBind.prototype.removeStaleListener = function(listener)
    {
        if (this.parentBind && this.parentModel && this.parentModel.isDeviceArray) {
            this.parentBind.removeStaleListener(listener);
        } else {
            gc.databind.AbstractBindValue.prototype.removeStaleListener.call(this, listener);
        }
    };

    gc.databind.internal.reg.RegisterBind.prototype.onValueChanged = function(oldValue, newValue, progress, forceWrite)
    {
        var isDeviceArray = this.parentModel && this.parentModel.isDeviceArray;
        if (this._q) {
            if (this.coreIndex === -1) {
                for (var i=0; i < newValue.length; i++) {
                    newValue[i] = Math.round(newValue[i] * Math.pow(2, this._q));
                }
            } else {
                if (!isNaN(newValue)) {
                    newValue = Math.round(newValue * Math.pow(2, this._q));
                }
            }
        }

        if (isDeviceArray) {
            this.parentBind.setValueFor(newValue, this.coreIndex, forceWrite);
        } else {
            this.setValue(newValue, progress, forceWrite);
        }
        if (newValue !== undefined) {
            this.notifyChangedListeners(oldValue, newValue, null);
        }
    };
    gc.databind.internal.reg.RegisterBind.prototype.setValue = function(newValue, progress, forceWrite) {
        if (this.isReadOnly()) {
            return;
        }
        var oldValue = this.fCachedValue;
        this.fCachedValue = newValue;
        if (!this.fDeferredMode) {
            if ((newValue !== undefined) && (forceWrite || (newValue !== oldValue))) {
                this.parentModel.writeValue(this.uri, this.fCachedValue, this.coreIndex).then(function () {
                    this.notifyChangedListeners(oldValue, newValue, null);
                }.bind(this)).fail(this._onFailure);
            }
        }
        return newValue;
    };
    // ...]

    // ------------------------------

    // [... Methods that are used only if model.isDeviceArray is false
    gc.databind.internal.reg.RegisterBind.prototype.writeValue = function(callback)
    {
        this._callback = callback;
        this.parentModel.writeValue(this.uri, this.fCachedValue, this.coreIndex).then(callback).fail(this._onFailure);
    };

    gc.databind.internal.reg.RegisterBind.prototype.readValue = function(callback)
    {
        this._callback = callback;
        this.parentModel.readValue(this.uri, this.coreIndex).then(this._onSuccess).then(callback).fail(this._onFailure);
    };

    gc.databind.internal.reg.RegisterBind.prototype.reportErrorStatus = function(errorMessage)
    {
        var status = null;
        if (errorMessage && errorMessage.length > 0)
        {
            status = gc.databind.AbstractStatus.createErrorStatus(errorMessage, 'target');
        }
        this.reportCriticalError(status);
    };
    // ...]
    gc.databind.internal.reg.RegisterBind.prototype.onRefresh = function(progress) {
        var isDeviceArray = this.parentModel && this.parentModel.isDeviceArray;
        if (isDeviceArray) {
            gc.databind.internal.reg.RegisterAllBind.prototype.onRefresh.call(this, progress);
        } else {
            gc.databind.AbstractAsyncBindValue.prototype.onRefresh.call(this, progress);
        }
    };

    gc.databind.internal.reg.RegisterBind.prototype.updateRegisterInfo = function(symbolData)
    {
        if (symbolData)
        {
            var defaultValue;
            if (this.parentBind && this.parentModel.isDeviceArray) {
                defaultValue = this.parentBind.defaultValue;
            } else {
                defaultValue = gc.utils.string2value(symbolData.value || symbolData['default']);
            }
            if (defaultValue !== undefined)  // restore default for the new device, before reading the actual value.
            {
                this.updateValue(defaultValue, undefined, true);
                this.setStale(false);
            }
            // support for qualifiers in register symbol data
            if (!this._readable)              // remove existing qualifiers
            {
                this._readable = true;
            }
            if (!this._writable)
            {
                this._writable = true;
            }
            if (!this._volatile)
            {
                this._volatile = true;
            }

            var type = symbolData.type;
            if (type)
            {
                // add any new qualifiers
                if (type === 'R')
                {
                    type = 'readonly';
                }
                else if (type === 'W')
                {
                    type = 'writeonly';
                }
                this.addQualifier(type);
            }

            // clear errors on successful read
            this.reportCriticalError(null);
        }
    };

    //----------------------------------

    var onChangedListener = function()
    {
        //var value = (this.parentBind.getValue() & this._mask) >>> this._shift;
        var value = gc.utils.getBitfieldValue(this.parentBind.getValue(), this._startBit, this._stopBit);

        if (this._maxValue && (value > this._maxValue))
        {
            // convert to signed
            value = value - (this._maxValue + 1);
        }
        if (this._q && !isNaN(value))
        {
            value = value / (Math.pow(2, this._q));
        }

        this.updateValue(value, undefined, true);
    };

    gc.databind.internal.reg.FieldBind = function(name, registerBind, symbolData, coreKey)
    {
        gc.databind.AbstractBindValue.call(this);

        this.uri = name;
        var parentChangedListener = new gc.databind.IChangedListener();
        this.parentBind = registerBind;
        var model = registerBind.parentModel;
        if (model && model.isDeviceArray) {
            if (!registerBind.parentBind) {
                if (registerBind.registerAllBind) {
                    registerBind.parentBind = registerBind.registerAllBind;
                }
            }
            this.coreIndex = parseCoreKey(coreKey);
            if (this.coreIndex < 0) {
                this.fCachedValue = [];
                this.fCommittedValue = [];
            }
            parentChangedListener.onValueChanged = onFieldParentChangedListener.bind(this);
        } else {
            parentChangedListener.onValueChanged = onChangedListener.bind(this);
        }
        parentChangedListener.onStatusChanged = onStatusListener.bind(this);
        registerBind.addChangedListener(parentChangedListener);
        registerBind.addStatusListener(parentChangedListener);

        this.dispose = function()
        {
            registerBind.removeChangedListener(parentChangedListener);
            registerBind.removeStatusListener(parentChangedListener);
        };

        this.updateRegisterInfo(symbolData);
    };

    gc.databind.internal.reg.FieldBind.prototype = new gc.databind.AbstractBindValue('number');

    gc.databind.internal.reg.FieldBind.prototype.excludeFromStorageProviderData = true;

    gc.databind.internal.reg.FieldBind.prototype._calcShiftMask = function(symbolData) {
        // setup default mask, shift and bitWidth for when no symbol data is available.
        var shift = 0;
        var mask = 1;
        var startBit = 0;
        var stopBit = 0;
        var regWidth = 8;
        var bitWidth = 1;
        if (symbolData) {
            startBit = gc.utils.string2value(symbolData.start);
            stopBit = gc.utils.string2value(symbolData.stop);
            if (symbolData.parentRegister) {
                regWidth = symbolData.parentRegister.size;
            }
            startBit = startBit || 0;
            stopBit = stopBit || startBit;
            bitWidth = stopBit - startBit + 1;
            mask = gc.utils.getMask(startBit,stopBit);
            shift = startBit;
        }

        return {shift: shift, mask: mask, bitWidth: bitWidth, regWidth: regWidth, startBit: startBit, stopBit: stopBit};
    };
    gc.databind.internal.reg.FieldBind.prototype.updateRegisterInfo = function(symbolData)
    {
        try
        {
            var tmp = this._calcShiftMask(symbolData);
            var bitWidth = tmp.bitWidth;
            this._mask = tmp.mask;
            this._shift = tmp.shift;
            this._startBit = +tmp.startBit;
            this._stopBit = +tmp.stopBit;
            this._regWidth = +tmp.regWidth;

            if (symbolData)
            {
                var type = symbolData.type;
                if (type)
                {
                    var match = typeParser.exec(type);
                    if (match && match.index === 0)
                    {
                        var isUnsigned = match[1] && match[1].toLowerCase() === 'unsigned';
                        if (!isUnsigned)
                        {
                            this._maxValue = gc.utils.getBitfieldMaxValue(bitWidth);
                        }
                        if (match[3] !== undefined) 
                        {
                            var q = +match[3];
                            if (q === undefined || isNaN(q) || q < 0)
                            {
                                throw "invalid type declaration for field: " + name;
                            }
                            else 
                            {
                                this._q = q;
                            }
                        }
                    }
                    else
                    {
                        throw "invalid type declaration for field: " + name;
                    }
                }

                // initialize value based on default register value.
                var model = this.parentBind.parentModel;
                if (model && model.isDeviceArray) {
                    onFieldParentChangedListener.call(this);
                } else {
                    onChangedListener.call(this);
                }
                this.fCachedStatus = undefined;
                this.setStatus(this.parentBind.getStatus());  // clear any errors,
            }
            else
            {
                throw 'Bit field "' + this.uri + '" is not recognized for this device.';
            }
        }
        catch(e)
        {
            this.fCachedStatus = gc.databind.AbstractStatus.createErrorStatus(e);
            this.setStatus(this.fCachedStatus);
        }
    };

    gc.databind.internal.reg.FieldBind.prototype.onValueChanged = function(oldValue, newValue, progress)
    {
        var model = this.parentBind.parentModel;
        var regValue;
        if (model && model.isDeviceArray && (this.coreIndex === -1)) {
            if (!Array.isArray(newValue)) {
                newValue = [newValue];
            } else {
                newValue = newValue.slice();
            }
            for (var i=0; i < newValue.length; i++) {
                var value = newValue[i];
                if (this._q && !isNaN(value)) {
                    value = Math.round(value * Math.pow(2, this._q));
                }
                regValue = this.parentBind.getValueFor(i);
                regValue = gc.utils.setBitfieldValue(regValue, this._startBit, this._stopBit, this._regWidth, value, false);
                this.parentBind.setValueFor(i,regValue);
            }
        } else {
            newValue = +newValue;
            if (this._q && !isNaN(newValue)) {
                newValue = Math.round(newValue * Math.pow(2, this._q));
            }

            regValue = this.parentBind.getValue();
            regValue = gc.utils.setBitfieldValue(regValue, this._startBit, this._stopBit, this._regWidth, newValue, false);
            this.parentBind.setValue(regValue);
        }

        if (model && model.isDeviceArray && (newValue !== undefined)) {
            this.notifyChangedListeners(oldValue, newValue, null);
        }
    };

    gc.databind.internal.reg.FieldBind.prototype.onFirstDataReceivedListenerAdded = function()
    {
        this.parentBind.addStreamingListener(this);
    };

    /**
     * Method called when the last streaming listener is removed from the list.
     * Derived classes can override this method to be notified for this event.
     */
    gc.databind.internal.reg.FieldBind.prototype.onLastDataReceivedListenerRemoved = function()
    {
        this.parentBind.removeStreamingListener(this);
    };

    gc.databind.internal.reg.FieldBind.prototype.onDataReceived = function()
    {
        this.notifyDataReceivedListeners(this.fCachedValue);
    };

    gc.databind.internal.reg.FieldBind.prototype.isStale = function()
    {
        return this.parentBind.isStale();
    };

    gc.databind.internal.reg.FieldBind.prototype.addStaleListener = function(listener)
    {
        this.parentBind.addStaleListener(listener);
    };

    gc.databind.internal.reg.FieldBind.prototype.removeStaleListener = function(listener)
    {
        this.parentBind.removeStaleListener(listener);
    };

}());


